#  Copyright (C) 2025
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
from dataclasses import dataclass
from typing import Any

from frozendict import frozendict
from http_tools import IncomingRequest
from init_helpers import convert_to_type

from .call_parameter import CallParameter
from .spec_parameter import SpecParameter


@dataclass(frozen=True, slots=True, kw_only=True)
class SecurityParameter(CallParameter):
    """Parameter using value from Security"""
    name: str

    async def get(self, incoming_request: IncomingRequest, security_kwargs: dict[str, Any]) -> Any:
        value = security_kwargs[self.name]
        return convert_to_type(self.schema, value)

    def get_spec_parameters(self) -> frozendict[SpecParameter, None]:
        return frozendict()



#  Copyright (C) 2025
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import abc

from frozendict import frozendict

from ..call_parameter import CallParameter
from ..spec_parameter import SpecParameter


class ParameterAggregation(CallParameter, abc.ABC):
    @abc.abstractmethod
    def get_spec_parameters(self) -> frozendict['SpecParameter', None]:
        pass



#  Copyright (C) 2025
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import functools
from abc import ABC
from dataclasses import field, dataclass
from typing import Mapping, ClassVar

from frozendict import frozendict
from init_helpers import Jsonable
from init_helpers import NoValue

from ..example import get_examples_as_dict, BaseExample
from ..maybe_referencable_resource import MaybeReferencableResource
from ..spec_resource import SpecResource, SpecRef


@dataclass(frozen=True, slots=True)
class BaseSchema(MaybeReferencableResource, ABC):
    spec_path: ClassVar[tuple[str]] = ['components', 'schemas']
    default: Jsonable | NoValue = field(default=NoValue, kw_only=True)
    description: str = field(default='', kw_only=True)
    examples: tuple[BaseExample, ...] | None = field(default=tuple(), kw_only=True)
    key: str | None = field(default=None, kw_only=True)
    enum: tuple[Jsonable] | None = field(default=None, kw_only=True)

    @property
    def has_default(self) -> bool:
        return self.default is not NoValue

    def get_spec_dependencies(self) -> frozenset['SpecResource']:
        return frozenset(self.examples) if self.examples else frozenset()

    def get_spec_dict(self, dependency_to_ref: Mapping['SpecResource', SpecRef]) -> frozendict[str, Jsonable]:
        result = {}
        result |= {'default': self.default} if self.default is not NoValue else {}
        result |= {'description': self.description} if self.description else {}
        result |= get_examples_as_dict(self.examples, dependency_to_ref)
        result |= {'enum': self.enum} if self.enum else {}
        return frozendict(result)

    def get_key(self) -> None:
        return self.key

    @functools.cache
    def _get_repr_parts(self) -> tuple[str, ...]:
        parts = []
        parts += [f'default={self.default!r}'] if self.default is not NoValue else []
        parts += [f'description={self.description!r}'] if self.description else []
        parts += [f'examples={self.examples!r}'] if self.examples else []
        parts += [f'key={self.key!r}'] if self.key else []
        parts += [f'enum={self.enum!r}'] if self.enum else []
        return tuple(parts)



#  Copyright (C) 2025
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import datetime
import json
from dataclasses import dataclass
from graphlib import TopologicalSorter
from typing import Callable, ClassVar, Any

import yaml
from init_helpers import AnyType, custom_dumps

from .info import Info
from .maybe_referencable_resource import MaybeReferencableResource
from .operation import Operation
from .schema.base_schema import BaseSchema
from .schema.build import SchemaBuilder
from .spec import Spec
from .spec_resource import SpecResource
from .spec_ref import SpecRef


class SpecGenerator:
    openapi_version: ClassVar[str] = "3.1.0"
    schema_builder: ClassVar[Callable[[AnyType], BaseSchema]] = SchemaBuilder.build_schema

    @staticmethod
    def resource_ref_condition(res: MaybeReferencableResource, parents: frozenset[SpecResource]) -> bool:
        return isinstance(res, BaseSchema) or len(parents) > 1

    @dataclass
    class Config:
        version: str = ''  # empty means autogenerate based on current datetime

    @dataclass
    class Context:
        company: str
        project_group: str
        project_name: str

    def __init__(
            self, config: Config, context: Context,
    ) -> None:
        self.version = config.version or self._get_version()
        self.config = config
        self.context = context
        self.info = Info(version=self.version,
                         title=f"{self.context.company}:{self.context.project_group}:{self.context.project_name}")
        self.resources: list[MaybeReferencableResource] = []

    def register(self, resource: MaybeReferencableResource) -> None:
        if isinstance(resource, Operation):
            if any(r.operation_id == r.operation_id for r in self.resources if isinstance(r, Operation)):
                KeyError(f"operation_id {resource.operation_id!r} already registered")
        self.resources.append(resource)

    def get_spec_dict(self, operation_filter: Callable[[Operation], bool] | None = None) -> dict:
        return self._get_spec(operation_filter=operation_filter).payload

    def get_spec_json(self, operation_filter: Callable[[Operation], bool] | None = None) -> bytes:
        return custom_dumps(self.get_spec_dict(operation_filter=operation_filter), indent=2)

    def get_spec_yaml(self, operation_filter: Callable[[Operation], bool] | None = None) -> bytes:
        return yaml.dump(json.loads(self.get_spec_json(operation_filter=operation_filter)), sort_keys=False).encode()

    def _wrap_non_spec_resource_element(self, element: Any) -> SpecResource:
        if isinstance(element, SpecResource | SpecRef):
            return element
        if isinstance(element, AnyType):
            return self.schema_builder(element)
        raise TypeError(f'Unsupported element type {type(element)=}, {element=}')

    def _get_spec(self, operation_filter: Callable[[Operation], bool] | None = None) -> Spec:
        resource_to_dependencies: dict[SpecResource, frozenset[SpecResource | SpecRef]] = {}
        pending: list[SpecResource] = (
            [r for r in self.resources if not isinstance(r, Operation) or operation_filter(r)]
            if operation_filter else self.resources[:])

        while pending:
            current: SpecResource | SpecRef = self._wrap_non_spec_resource_element(pending.pop())
            if isinstance(current, SpecRef):
                continue
            if current in resource_to_dependencies:
                continue

            current_dependencies: frozenset[SpecResource | SpecRef] = frozenset(
                self._wrap_non_spec_resource_element(dep) for dep in current.get_spec_dependencies()
            )
            resource_to_dependencies[current] = current_dependencies
            pending.extend(current_dependencies)

        resource_to_parents: dict[SpecResource, set[SpecResource | SpecRef]] = {}
        for resource, dependencies in resource_to_dependencies.items():
            for dependency in dependencies:
                resource_to_parents.setdefault(dependency, set()).add(resource)

        result = Spec(self.info, schema_builder=self._wrap_non_spec_resource_element)

        dependency_spec_refs: set[SpecRef] = set()
        for resource in TopologicalSorter(resource_to_dependencies).static_order():
            if isinstance(resource, SpecRef):
                dependency_spec_refs.add(resource)
                continue
            do_add = False
            key = None
            if isinstance(resource, MaybeReferencableResource) and (key := resource.get_key()):
                parents = resource_to_parents.get(resource, frozenset())
                do_add = not parents or not resource.can_be_inlined or self.resource_ref_condition(resource, parents)

            # print(resource)
            # print()
            if key and do_add:
                result.add_resource(resource, key)
            else:
                result.prepare_resource(resource)
        for ref in dependency_spec_refs:
            assert result.has_ref(ref)
        # print()
        return result

    @staticmethod
    def _get_version() -> str:
        now = datetime.datetime.now()
        return f"{now.year - 2021}.{now.month}.{now.day}{now.hour:02d}{now.minute:02d}"




#  Copyright (C) 2025
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
from .body_parameter import BodyParameter
from .json_body_parameter import JsonBodyParameter
from .json_whole_body_parameter import JsonWholeBodyParameter
from .raw_body_parameter import RawBodyParameter



#  Copyright (C) 2025
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
from dataclasses import dataclass
from typing import ClassVar

from .type_schema import TypeSchema


@dataclass(frozen=True, slots=True)
class NullSchema(TypeSchema):
    type_: ClassVar[str] = "null"

    def __repr__(self):
        return f'{self.__class__.__name__}({", ".join(self._get_repr_parts())})'

    __str__ = __repr__




#  Copyright (C) 2025
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import functools
from dataclasses import dataclass
from typing import Mapping

import yarl
from frozendict import frozendict
from init_helpers import Jsonable, AnyType

from .spec_ref import SpecRef
from .spec_resource import SpecResource


@dataclass(frozen=True, slots=True)
class Info(SpecResource):
    """Info https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.1.0.md#info-object"""
    title: str
    version: str
    summary: str = ''
    description: str = ''
    terms_of_service: yarl.URL | None = None
    # contact: Contact  # TODO: implement
    # license: License  # TODO: implement

    def get_spec_dependencies(self) -> frozenset[SpecResource]:
        return frozenset()

    def get_spec_dict(self, dependency_to_ref: Mapping[SpecResource | AnyType, SpecRef]) -> frozendict[str, Jsonable]:
        result = {'title': self.title, 'version': self.version}
        result |= {"summary": self.summary} if self.summary else {}
        result |= {"description": self.description} if self.description else {}
        result |= {"terms_of_service": str(self.terms_of_service)} if self.terms_of_service else {}
        return frozendict(result)

    @functools.cache
    def _get_repr_parts(self) -> tuple[str, ...]:
        parts = [f'title={self.title!r}', f'version={self.version!r}']
        parts += [f'summary={self.summary!r}'] if self.summary else []
        parts += [f'description={self.description!r}'] if self.description else []
        parts += [f'terms_of_service={str(self.terms_of_service)!r}'] if self.terms_of_service else []
        return tuple(parts)

    def __repr__(self):
        return f'{self.__class__.__name__}({", ".join(self._get_repr_parts())})'

    __str__ = __repr__



#  Copyright (C) 2025
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import abc
import functools
import json
from dataclasses import dataclass, field
from typing import Callable, Mapping, Iterable, ClassVar

import http_tools
import yarl
from frozendict import frozendict
from init_helpers import custom_dumps, Jsonable, DataclassProtocol
from more_itertools import first

from .spec_ref import SpecRef
from .spec_resource import SpecResource


@dataclass(frozen=True, slots=True)
class BaseExample(SpecResource, abc.ABC):
    name: str = field(default='', kw_only=True)
    summary: str = field(default='', kw_only=True)
    description: str = field(default='', kw_only=True)

    def __init__(self, *, name: str = None, summary: str = '', description: str = ''):
        object.__setattr__(self, 'name', name)
        object.__setattr__(self, 'summary', summary)
        object.__setattr__(self, 'description', description)

    def get_spec_dependencies(self) -> frozenset[SpecResource]:
        return frozenset()

    @abc.abstractmethod
    def get_spec_dict(self, dependency_to_ref: Mapping['SpecResource', SpecRef]) -> frozendict[str, Jsonable]:
        result = {}
        result |= {'summary': self.summary} if self.summary else {}
        result |= {'description': self.description} if self.description else {}
        return frozendict(result)

    def get_name(self, i: int):
        return self.name or f'Example {i}'

    @functools.cache
    def _get_repr_parts(self) -> tuple[str, ...]:
        result = []
        result += [f'name={self.name!r}'] if self.name else []
        result += [f'summary={self.summary!r}'] if self.summary else []
        result += [f'description={self.description!r}'] if self.description else []
        return tuple(result)

    def __repr__(self):
        return f'{self.__class__.__name__}({", ".join(self._get_repr_parts())})'

    __str__ = __repr__


@dataclass(frozen=True, slots=True)
class Example(BaseExample):
    jsoner: ClassVar[Callable[[Jsonable], str]] = staticmethod(custom_dumps)
    jsoned: str = field(init=False)

    def __init__(self, val: Jsonable | DataclassProtocol, *, name: str = None, summary: str = '', description: str = ''):
        kwargs = {}
        kwargs |= {'name': name} if name else {}
        kwargs |= {'summary': summary} if summary else {}
        kwargs |= {'description': description} if description else {}
        BaseExample.__init__(self, **kwargs)
        jsoned = self.jsoner(val)
        object.__setattr__(self, 'jsoned', jsoned)

    def get_spec_dict(self, dependency_to_ref: Mapping['SpecResource', SpecRef]) -> frozendict[str, Jsonable]:
        result = {"value": self.value}
        return frozendict(BaseExample.get_spec_dict(self, dependency_to_ref) | result)

    @property
    def value(self) -> Jsonable:
        return json.loads(self.jsoned)

    def _get_repr_parts(self) -> tuple[str, ...]:
        return (f"value={self.value!r}", ) + BaseExample._get_repr_parts(self)

    def __repr__(self):
        return f'{self.__class__.__name__}({", ".join(self._get_repr_parts())})'

    __str__ = __repr__

# @dataclass(frozen=True)
# class RawExample(Example):
#     def __post_init__(self, val: Jsonable):
#         object.__setattr__(self, 'jsoned', val)


@dataclass(frozen=True, slots=True)
class AnswerExample(Example):
    def __init__(self, answer: http_tools.Answer, *, name: str = None, summary: str = '', description: str = ''):
        kwargs = {}
        kwargs |= {'name': name} if name else {}
        kwargs |= {'summary': summary} if summary else {}
        kwargs |= {'description': description} if description else {}
        Example.__init__(self, json.loads(answer.body.decode()), **kwargs)
        if answer.content_type != http_tools.ContentType.Json:
            raise TypeError(f"Only Json is supported, got: {answer.content_type}")

    def _get_repr_parts(self) -> tuple[str, ...]:
        return (f"value={json.dumps(self.value)}", ) + BaseExample._get_repr_parts(self)

    def __repr__(self):
        return f'{self.__class__.__name__}({", ".join(self._get_repr_parts())})'

    __str__ = __repr__


@dataclass(frozen=True, slots=True)
class UrlExample(BaseExample):
    url: yarl.URL

    def get_spec_dict(self, dependency_to_ref: Mapping['SpecResource', SpecRef]) -> frozendict[str, Jsonable]:
        result = {"externalValue": str(self.url)}
        return frozendict(BaseExample.get_spec_dict(self, dependency_to_ref) | result)

    @functools.cache
    def _get_repr_parts(self) -> tuple[str, ...]:
        return (f"url={self.url!r}", ) + BaseExample._get_repr_parts(self)

    def __repr__(self):
        return f'{self.__class__.__name__}({", ".join(self._get_repr_parts())})'

    __str__ = __repr__


def get_examples_as_dict(
        examples: Iterable[BaseExample],
        dependency_to_ref: Mapping['SpecResource', SpecRef]
) -> frozendict[str, Jsonable]:
    result = {}
    if examples:
        if example := first((e for e in examples if isinstance(e, Example)), default=None):
            result['example'] = example.value
        result['examples'] = {e.get_name(i): dependency_to_ref[e] for i, e in enumerate(examples, 1)}
    return frozendict(result)



#  Copyright (C) 2025
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
from .security_requirement import SecurityRequirement
